/******************************
 * 共有メモリイメージを作成
 ******************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <wayland-client.h>

//--------------

struct wl_shm *g_shm;
unsigned int g_shm_cnt = 0; //共有メモリ名用

typedef struct
{
	struct wl_shm_pool *pool;
	struct wl_buffer *buffer;
	void *data;
	int size;
}imagebuf;

//-------------


/* POSIX 共有メモリオブジェクトを作成
 *
 * return: shm_open() の戻り値 (成功時はファイルディスクリプタ) */

static int _create_posix_shm()
{
	char name[64];
	int ret;

	while(1)
	{
		snprintf(name, 64, "/wayland-test-%d", g_shm_cnt);
		
		ret = shm_open(name, O_CREAT | O_EXCL | O_RDWR | O_CLOEXEC, 0600);

		if(ret >= 0)
		{
			//成功
			
			shm_unlink(name);

			g_shm_cnt++;
			break;
		}
		else if(errno == EEXIST)
			//同名が存在する -> カウントを+1して次へ
			g_shm_cnt++;
		else if(errno != EINTR)
		{
			printf("failed shm_open()\n");
			break;
		}
	}

	return ret;
}

/* wl_shm_pool 作成
 *
 * size: バッファサイズ
 * ppbuf: バッファのポインタが入る */

static struct wl_shm_pool *_create_shm_pool(int size,void **ppbuf)
{
	int fd;
	void *data;
	struct wl_shm_pool *pool;

	fd = _create_posix_shm();
	if(fd < 0) return NULL;

	printf("posix shm: %d\n", fd);

	//サイズ変更

	if(ftruncate(fd, size) < 0)
	{
		close(fd);
		return NULL;
	}

	//メモリにマッピング

	data = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);

	if(data == MAP_FAILED)
	{
		close(fd);
		return NULL;
	}

	printf("mmap: %p\n", data);

	//wl_shm_pool 作成

	pool = wl_shm_create_pool(g_shm, fd, size);

	//クローズしてもアンマップはされない
	close(fd);

	*ppbuf = data;

	return pool;
}

/* イメージバッファ作成 */

static imagebuf *imagebuf_create(int width,int height)
{
	imagebuf *img;
	struct wl_shm_pool *pool;
	struct wl_buffer *buffer;
	void *buf;
	int size;

	size = width * 4 * height;

	//wl_shm_pool 作成

	pool = _create_shm_pool(size, &buf);
	if(!pool) return NULL;

	printf("create shm pool: ok\n");

	//wl_buffer 作成

	buffer = wl_shm_pool_create_buffer(pool,
		0, //offset
		width, height,
		width * 4, //Y1行のバイト数
		WL_SHM_FORMAT_XRGB8888);

	if(buffer)
		printf("create buffer: ok\n");

	//imagebuf

	img = (imagebuf *)malloc(sizeof(imagebuf));
	if(!img) return NULL;

	img->pool = pool;
	img->buffer = buffer;
	img->data = buf;
	img->size = size;

	return img;
}

/* imagebuf 削除 */

static void imagebuf_destroy(imagebuf *p)
{
	if(p)
	{
		wl_buffer_destroy(p->buffer);
		wl_shm_pool_destroy(p->pool);
		munmap(p->data, p->size);
		
		free(p);
	}
}


//---------------------
// wl_registry
//---------------------


static void _registry_global(
	void *data,struct wl_registry *registry,
	uint32_t id,const char *interface,uint32_t version)
{
	//wl_shm をバインド
	
	if(strcmp(interface, "wl_shm") == 0)
		g_shm = wl_registry_bind(registry, id, &wl_shm_interface, 1);
}

static void _registry_global_remove(void *data,struct wl_registry *registry,uint32_t id)
{

}

static const struct wl_registry_listener g_reg_listener = {
	_registry_global, _registry_global_remove
};


//------------


int main(void)
{
	struct wl_display *display;
	struct wl_registry *reg;
	imagebuf *img;

	display = wl_display_connect(NULL);
	if(!display) return 1;

	//wl_registry

	reg = wl_display_get_registry(display);
	wl_registry_add_listener(reg, &g_reg_listener, NULL);

	wl_display_roundtrip(display);

	//イメージ作成

	img = imagebuf_create(100, 100);
	
	imagebuf_destroy(img);

	//

	wl_shm_destroy(g_shm);
	wl_registry_destroy(reg);

	wl_display_disconnect(display);

	return 0;
}
