/******************************
 * xdg-shell でウィンドウ表示
 ******************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <wayland-client.h>
#include "xdg-shell-client-protocol.h"

#include "client.h"
#include "imagebuf.h"


//-----------------

/* クライアント */

typedef struct
{
	Client b;

	struct xdg_wm_base *xdg_wm_base;
}ClientEx;

/* ウィンドウ */

typedef struct
{
	ClientEx *client;
	struct wl_surface *surface;
	struct xdg_surface *xdg_surface;
	struct xdg_toplevel *xdg_toplevel;

	ImageBuf *img;
	uint8_t flag_configure;
}WindowXdg;

void WindowXdg_update(WindowXdg *p);

//-----------------


//--------------------
// xdg_surface
//--------------------


static void _xdg_surface_configure(
	void *data,struct xdg_surface *surface,uint32_t serial)
{
	WindowXdg *win = (WindowXdg *)data;

	printf("xdg_surface#configure | serial %u\n", serial);

	//commit の前に実行

	xdg_surface_ack_configure(surface, serial);

	//初期表示時、ウィンドウを更新

	if(win->flag_configure == 0)
	{
		win->flag_configure = 1;

		ImageBuf_fill(win->img, 0xffff0000);

		WindowXdg_update(win);
	}
}

static const struct xdg_surface_listener g_xdg_surface_listener = {
	_xdg_surface_configure
};


//--------------------
// xdg_toplevel
//--------------------


/* サイズ変更時や状態変更時 */

static void _xdg_toplevel_configure(
	void *data, struct xdg_toplevel *toplevel,
	int32_t width, int32_t height, struct wl_array *states)
{
	uint32_t *ps;

	printf("xdg_toplevel#configure | w:%d, h:%d, states:", width, height);

	wl_array_for_each(ps, states)
	{
		switch(*ps)
		{
			case XDG_TOPLEVEL_STATE_MAXIMIZED:
				printf("MAXIMIZED ");
				break;
			case XDG_TOPLEVEL_STATE_FULLSCREEN:
				printf("FULLSCREEN ");
				break;
			case XDG_TOPLEVEL_STATE_RESIZING:
				printf("RESIZING ");
				break;
			case XDG_TOPLEVEL_STATE_ACTIVATED:
				printf("ACTIVATED ");
				break;
		}
	}

	printf("\n");
}

/* ユーザーによって、ウィンドウを閉じる要求が行われた時 */

static void _xdg_toplevel_close(void *data,struct xdg_toplevel *toplevel)
{
	printf("xdg_toplevel#close\n");

	//終了させる

	((WindowXdg *)data)->client->b.finish_loop = 1;
}

const struct xdg_toplevel_listener g_xdg_toplevel_listener = {
	_xdg_toplevel_configure, _xdg_toplevel_close
};


//------------------------
// WindowXdg
//------------------------


/* ウィンドウ作成 */

WindowXdg *WindowXdg_create(ClientEx *cl,int width,int height)
{
	WindowXdg *p;

	p = (WindowXdg *)calloc(1, sizeof(WindowXdg));
	if(!p) return NULL;

	p->client = cl;

	//wl_surface

	p->surface = wl_compositor_create_surface(cl->b.compositor);

	//xdg_surface

	p->xdg_surface = xdg_wm_base_get_xdg_surface(cl->xdg_wm_base, p->surface);

	xdg_surface_add_listener(p->xdg_surface, &g_xdg_surface_listener, p);

	//xdg_toplevel

	p->xdg_toplevel = xdg_surface_get_toplevel(p->xdg_surface);
	
	xdg_toplevel_add_listener(p->xdg_toplevel, &g_xdg_toplevel_listener, p);

	//適用

	wl_surface_commit(p->surface);

	//イメージ作成

	p->img = ImageBuf_create(cl->b.shm, width, height);

	return p;
}

/* ウィンドウ破棄 */

void WindowXdg_destroy(WindowXdg *p)
{
	if(p)
	{
		ImageBuf_destroy(p->img);
	
		xdg_toplevel_destroy(p->xdg_toplevel);
		xdg_surface_destroy(p->xdg_surface);
		wl_surface_destroy(p->surface);
	
		free(p);
	}
}

/* ウィンドウ更新 */

void WindowXdg_update(WindowXdg *p)
{
	wl_surface_attach(p->surface, p->img->buffer, 0, 0);
	wl_surface_damage(p->surface, 0, 0, p->img->width, p->img->height);
	wl_surface_commit(p->surface);
}


//-------------------------
// xdg_wm_base
//-------------------------


static void _xdg_wm_base_ping(void *data,
	struct xdg_wm_base *xdg_wm_base,uint32_t serial)
{
	//応答する
	xdg_wm_base_pong(xdg_wm_base, serial);
}

const struct xdg_wm_base_listener g_xdg_wm_base_listener = {
	_xdg_wm_base_ping
};


static void _registry_global(void *data,
	struct wl_registry *reg,uint32_t id,const char *name,uint32_t ver)
{
	ClientEx *p = (ClientEx *)data;

	if(strcmp(name, "xdg_wm_base") == 0)
	{
		p->xdg_wm_base = wl_registry_bind(reg, id, &xdg_wm_base_interface, 1);

		xdg_wm_base_add_listener(p->xdg_wm_base, &g_xdg_wm_base_listener, NULL);
	}
}

	
//--------------


int main(void)
{
	ClientEx *client;
	WindowXdg *win;

	client = (ClientEx *)Client_new(sizeof(ClientEx));

	client->b.registry_global = _registry_global;

	Client_init(CLIENT(client));

	//サーバーでインターフェイスが存在しなかった

	if(!client->xdg_wm_base)
	{
		printf("not fount 'xdg_wm_base'\n");
		Client_destroy(CLIENT(client));
		return 1;
	}

	//ウィンドウ作成

	win = WindowXdg_create(client, 256, 256);

	//ここでウィンドウを更新しようとすると、エラーになる

#if 0
	ImageBuf_fill(win->img, 0xffff0000);

	WindowXdg_update(win);
#endif

	//イベントループ

	Client_loop_simple(CLIENT(client));

	//

	WindowXdg_destroy(win);

	xdg_wm_base_destroy(client->xdg_wm_base);

	Client_destroy(CLIENT(client));

	return 0;
}
