/******************************
 * ポインタデバイス入力
 ******************************/

#include <stdio.h>
#include <string.h>
#include <linux/input.h>

#include <wayland-client.h>

#include "client.h"
#include "imagebuf.h"


//-------------

typedef struct
{
	Client b;

	struct wl_seat *seat;
	struct wl_pointer *pointer;
	uint32_t seat_ver;
}ClientEx;

#define _FLUSH   fflush(stdout)

//-------------


/* wl_pointer 解放 */

static void _pointer_release(ClientEx *p)
{
	if(p->seat_ver >= WL_POINTER_RELEASE_SINCE_VERSION)
		wl_pointer_release(p->pointer);
	else
		wl_pointer_destroy(p->pointer);

	p->pointer = NULL;
}


//------------------------
// wl_pointer
//------------------------


/* ポインタがサーフェス内に入った時 */

static void _pointer_enter(void *data, struct wl_pointer *pointer,
	uint32_t serial, struct wl_surface *surface, wl_fixed_t x, wl_fixed_t y)
{
	printf("enter | (%.2f, %.2f)\n", x / 256.0, y / 256.0);
	_FLUSH;
}

/* ポインタがサーフェス外に出た時 */

static void _pointer_leave(void *data, struct wl_pointer *pointer,
	uint32_t serial, struct wl_surface *surface)
{
	printf("leave\n");
	_FLUSH;
}

/* サーフェス内でポインタが移動した時 */

static void _pointer_motion(void *data, struct wl_pointer *pointer,
	uint32_t time, wl_fixed_t x, wl_fixed_t y)
{
	printf("motion | (%.2f, %.2f), time:%u\n",
		x / 256.0, y / 256.0, time);

	_FLUSH;
}

/* ボタンが押された時と離された時 */

static void _pointer_button(void *data, struct wl_pointer *pointer,
	uint32_t serial, uint32_t time, uint32_t button, uint32_t state)
{
	printf("button | %s, bttno:0x%X, time:%u\n",
		(state == WL_POINTER_BUTTON_STATE_PRESSED)? "press": "release",
		button, time);

	_FLUSH;

	//中ボタンで終了

	if(button == BTN_MIDDLE)
		((Client *)data)->finish_loop = 1;
}

/* ホイールスクロール、または他の軸 */

static void _pointer_axis(void *data, struct wl_pointer *pointer,
	uint32_t time, uint32_t axis, wl_fixed_t value)
{
	printf("axis | %s, value:%.2f, time:%u\n",
		(axis == WL_POINTER_AXIS_VERTICAL_SCROLL)? "vert": "horz",
		value / 256.0, time);

	_FLUSH;
}

//===== 以下は、ver 5 から

/* 属している一連のイベントの終了時 */

static void _pointer_frame(void *data, struct wl_pointer *pointer)
{
	printf("frame\n");
	_FLUSH;
}

/* 軸のソース情報 (frame イベントの前) */

static void _pointer_axis_source(void *data, struct wl_pointer *pointer, uint32_t axis_source)
{
	printf("axis_source | %u\n", axis_source);
	_FLUSH;
}

/* 軸の通知の停止 */

static void _pointer_axis_stop(void *data, struct wl_pointer *pointer, uint32_t time, uint32_t axis)
{
	printf("axis_stop | axis:%u, time:%u\n", axis, time);
	_FLUSH;
}

/* 軸のステップ情報 */

static void _pointer_axis_discrete(void *data, struct wl_pointer *pointer, uint32_t axis, int32_t discrete)
{
	printf("axis_discrete | axis:%u, discrete:%d\n", axis, discrete);
	_FLUSH;
}

static const struct wl_pointer_listener g_pointer_listener = {
	_pointer_enter, _pointer_leave, _pointer_motion, _pointer_button,
	_pointer_axis, _pointer_frame, _pointer_axis_source,
	_pointer_axis_stop, _pointer_axis_discrete
};


//------------------------
// wl_seat
//------------------------


/* デバイスの利用状態が変更された時 */

static void _seat_capabilities(void *data, struct wl_seat *seat, uint32_t cap)
{
	ClientEx *p = (ClientEx *)data;

	printf("wl_seat # capabilities | cap:0x%X\n", cap);

	if((cap & WL_SEAT_CAPABILITY_POINTER) && !p->pointer)
	{
		//wl_pointer 作成

		p->pointer = wl_seat_get_pointer(seat);

		wl_pointer_add_listener(p->pointer, &g_pointer_listener, p);

	}
	else if(!(cap & WL_SEAT_CAPABILITY_POINTER) && p->pointer)
	{
		//wl_pointer 解放

		_pointer_release(p);
	}
}

static void _seat_name(void *data, struct wl_seat *wl_seat, const char *name)
{
	printf("wl_seat # name: \"%s\"\n", name);
}

static const struct wl_seat_listener g_seat_listener = {
	_seat_capabilities, _seat_name
};


//------------------------


static void _registry_global(
    void *data,struct wl_registry *reg,uint32_t id,const char *name,uint32_t ver)
{
	ClientEx *p = (ClientEx *)data;

	if(strcmp(name, "wl_seat") == 0)
	{
		if(ver >= 5) ver = 5;

		p->seat = wl_registry_bind(reg, id, &wl_seat_interface, ver);
		p->seat_ver = ver;

		wl_seat_add_listener(p->seat, &g_seat_listener, p);
	}
}

/* ClientEx 解放 */

static void _clientex_destroy(Client *cl)
{
	ClientEx *p = (ClientEx *)cl;

	//wl_pointer

	_pointer_release(p);

	//wl_seat

	if(p->seat_ver >= WL_SEAT_RELEASE_SINCE_VERSION)
		wl_seat_release(p->seat);
	else
		wl_seat_destroy(p->seat);
}


//---------------


int main(void)
{
	ClientEx *p;
	Window *win;

	p = (ClientEx *)Client_new(sizeof(ClientEx));

	p->b.destroy = _clientex_destroy;
	p->b.registry_global = _registry_global;
	
	Client_init(CLIENT(p));

	//ウィンドウ

	win = Window_create(CLIENT(p), 256, 256, NULL);

	ImageBuf_fill(win->img, 0xffff0000);
	Window_update(win);

	//

	Client_loop_simple(CLIENT(p));

	//解放

	Window_destroy(win);

	Client_destroy(CLIENT(p));

	return 0;
}
