/******************************
 * カーソル形状の変更
 *
 * 中ボタン押しで終了。
 ******************************/

#include <stdio.h>
#include <string.h>
#include <linux/input.h>

#include <wayland-client.h>
#include <wayland-cursor.h>

#include "client.h"
#include "imagebuf.h"


//-------------

typedef struct
{
	Client b;
	
	struct wl_cursor_theme *cursor_theme; //カーソルテーマ
	struct wl_cursor *cursor[2];	//各カーソルデータ
	struct wl_surface *surface_cursor;	//カーソル用サーフェス
	struct wl_callback *callback;	//アニメコールバック

	uint32_t time_start,	//アニメ開始時間
		serial_enter; 		//enter 時の serial
	int cursor_no;			//現在のカーソル番号
}ClientEx;

//-------------

static void _surface_cursor_frame_callback(
	void *data,struct wl_callback *callback,uint32_t time);

//-------------



//----------------------
// カーソル変更
//----------------------


/* カーソル画像変更
 *
 * index: 複数枚ある場合、画像のインデックス */

static void _set_cursor_image(ClientEx *p,int index)
{
	struct wl_buffer *buffer;
	struct wl_cursor_image *img;

	img = p->cursor[p->cursor_no]->images[index];

	//wl_buffer 取得

	buffer = wl_cursor_image_get_buffer(img);
	if(!buffer) return;

	//

	wl_surface_attach(p->surface_cursor, buffer, 0, 0);
	wl_surface_damage(p->surface_cursor, 0, 0, img->width, img->height);
	wl_surface_commit(p->surface_cursor);

	//ポインタのカーソル形状としてセット
	//[!] ここで、enter 時の serial が必要

	wl_pointer_set_cursor(p->b.pointer, p->serial_enter,
		p->surface_cursor,
		img->hotspot_x, img->hotspot_y);
}

/* アニメーションコールバック */

static const struct wl_callback_listener g_cursor_frame_listener = {
	_surface_cursor_frame_callback
};

void _surface_cursor_frame_callback(
	void *data,struct wl_callback *callback,uint32_t time)
{
	ClientEx *p = (ClientEx *)data;
	int index,commit = 1;

	wl_callback_destroy(callback);

	//コールバック再セット

	p->callback = wl_surface_frame(p->surface_cursor);

	wl_callback_add_listener(p->callback, &g_cursor_frame_listener, p);

	//

	if(p->time_start == 0)
		//最初のコールバックの場合、開始時間をセット
		p->time_start = time;
	else
	{
		//2回目以降の場合、経過時間によって必要な画像を取得し、セット
	
		index = wl_cursor_frame(p->cursor[p->cursor_no], time - p->time_start);

		_set_cursor_image(p, index);

		commit = 0;
	}

	//コールバックを適用
	//[!] カーソル画像が変わった時はすでに実行しているので呼ばない

	if(commit)
		wl_surface_commit(p->surface_cursor);
}

/* カーソル形状変更 */

static void _change_cursor(ClientEx *p)
{
	struct wl_cursor *cursor;

	p->cursor_no ^= 1;

	cursor = p->cursor[p->cursor_no];

	if(cursor)
	{
		//wl_callback 破棄
	
		if(p->callback)
		{
			wl_callback_destroy(p->callback);
			p->callback = NULL;
		}
		
		//カーソル画像セット
		
		p->time_start = 0;

		_set_cursor_image(p, 0);

		//複数枚ある場合、アニメ開始

		if(cursor->image_count > 1)
		{
			p->callback = wl_surface_frame(p->surface_cursor);

			wl_callback_add_listener(p->callback, &g_cursor_frame_listener, p);

			wl_surface_commit(p->surface_cursor);

			printf("start animation: %d images\n", cursor->image_count);
		}
	}
}


//----------------------
// wl_pointer
//----------------------


/* enter */

static void _pointer_enter(void *data, struct wl_pointer *pointer,
	uint32_t serial, struct wl_surface *surface, wl_fixed_t x, wl_fixed_t y)
{
	ClientEx *p = (ClientEx *)data;

	p->serial_enter = serial;

	//カーソル変更

	_change_cursor(p);

	printf("change cursor: %d\n", p->cursor_no);
}

static void _pointer_leave(void *data, struct wl_pointer *pointer,
	uint32_t serial, struct wl_surface *surface)
{

}

static void _pointer_motion(void *data, struct wl_pointer *pointer,
	uint32_t time, wl_fixed_t x, wl_fixed_t y)
{

}

/* ボタン */

static void _pointer_button(void *data, struct wl_pointer *pointer,
	uint32_t serial, uint32_t time, uint32_t button, uint32_t state)
{
	//中ボタンで終了

	if(button == BTN_MIDDLE
		&& state == WL_POINTER_BUTTON_STATE_PRESSED)
		CLIENT(data)->finish_loop = 1;
}

static void _pointer_axis(void *data, struct wl_pointer *pointer,
	uint32_t time, uint32_t axis, wl_fixed_t value)
{

}

static void _pointer_frame(void *data, struct wl_pointer *pointer)
{

}

static void _pointer_axis_source(void *data, struct wl_pointer *pointer, uint32_t axis_source)
{

}

static void _pointer_axis_stop(void *data, struct wl_pointer *pointer, uint32_t time, uint32_t axis)
{

}

static void _pointer_axis_discrete(void *data, struct wl_pointer *pointer, uint32_t axis, int32_t discrete)
{

}

static const struct wl_pointer_listener g_pointer_listener = {
	_pointer_enter, _pointer_leave, _pointer_motion, _pointer_button,
	_pointer_axis, _pointer_frame, _pointer_axis_source,
	_pointer_axis_stop, _pointer_axis_discrete
};


//------------------------


/* ClientEx 解放 */

static void _clientex_destroy(Client *cl)
{
	ClientEx *p = (ClientEx *)cl;

	wl_cursor_theme_destroy(p->cursor_theme);

	if(p->callback)
		wl_callback_destroy(p->callback);

	wl_surface_destroy(p->surface_cursor);
}

/* カーソル初期化 */

static void _cursor_init(ClientEx *p)
{
	//wl_surface 作成

	p->surface_cursor = wl_compositor_create_surface(p->b.compositor);

	//カーソルテーマ読み込み

	p->cursor_theme = wl_cursor_theme_load(NULL, 32, p->b.shm);

	//カーソル読み込み

	p->cursor[0] = wl_cursor_theme_get_cursor(p->cursor_theme, "text");
	p->cursor[1] = wl_cursor_theme_get_cursor(p->cursor_theme, "wait");

	if(!p->cursor[0]) printf("not found cursor 'text'\n");
	if(!p->cursor[1]) printf("not found cursor 'wait'\n");
}


//-----------------


int main(void)
{
	ClientEx *p;
	Window *win;

	p = (ClientEx *)Client_new(sizeof(ClientEx));

	p->b.destroy = _clientex_destroy;
	p->b.init_flags = INIT_FLAGS_SEAT | INIT_FLAGS_POINTER;
	p->b.pointer_listener = &g_pointer_listener;

	Client_init(CLIENT(p));

	//カーソル初期化

	_cursor_init(p);
	
	//ウィンドウ

	win = Window_create(CLIENT(p), 256, 256, NULL);

	ImageBuf_fill(win->img, 0xffff0000);
	Window_update(win);

	//

	Client_loop_simple(CLIENT(p));

	//解放

	Window_destroy(win);

	Client_destroy(CLIENT(p));

	return 0;
}
