/******************************
 * キーボード入力
 *
 * ESC キーで終了。
 ******************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <wayland-client.h>

#include <xkbcommon/xkbcommon.h>

#include "client.h"
#include "imagebuf.h"


//-------------

typedef struct
{
	Client b;

	struct wl_keyboard *keyboard;
	struct xkb_keymap *xkb_keymap;
	struct xkb_state *xkb_state;
}ClientEx;

//-------------


/* キーボード関連解放 */

static void _keyboard_release(ClientEx *p)
{
	//XKB 解放

	xkb_keymap_unref(p->xkb_keymap);
	xkb_state_unref(p->xkb_state);

	p->xkb_keymap = NULL;
	p->xkb_state = NULL;

	//wl_keyboard

	if(p->b.seat_ver >= WL_KEYBOARD_RELEASE_SINCE_VERSION)
		wl_keyboard_release(p->keyboard);
	else
		wl_keyboard_destroy(p->keyboard);

	p->keyboard = NULL;
}

/* XKB キーマップ作成 */

static void _xkb_keymap(ClientEx *p,char *mapstr)
{
	struct xkb_context *context;
	struct xkb_keymap *keymap;

	//コンテキスト作成

	context = xkb_context_new(XKB_CONTEXT_NO_FLAGS);
	if(!context) return;

	//キーマップ作成

	keymap = xkb_keymap_new_from_string(context,
			mapstr, XKB_KEYMAP_FORMAT_TEXT_V1, 0);

	if(!keymap)
	{
		xkb_context_unref(context);
		return;
	}

	//

	xkb_keymap_unref(p->xkb_keymap);
	xkb_state_unref(p->xkb_state);

	p->xkb_keymap = keymap;
	p->xkb_state = xkb_state_new(keymap);

	//

	xkb_context_unref(context);
}


//-----------------------
// wl_keyboard
//-----------------------


/* キーマップ */

static void _keyboard_keymap(void *data, struct wl_keyboard *keyboard,
	uint32_t format, int32_t fd, uint32_t size)
{
	char *mapstr;

	printf("keymap | format:%u, fd:%d, size:%u\n", format, fd, size);

	//XKB ver 1 以外は対応しない

	if(format != WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1)
	{
		close(fd);
		return;
	}

	//文字列としてメモリにマッピング

	mapstr = mmap(NULL, size, PROT_READ, MAP_SHARED, fd, 0);

	if(mapstr == MAP_FAILED)
	{
		close(fd);
		return;
	}

	//キーマップ作成

	_xkb_keymap((ClientEx *)data, mapstr);

	//

	munmap(mapstr, size);
	close(fd);
}

/* キーボードフォーカスが来た時 */

static void _keyboard_enter(void *data, struct wl_keyboard *keyboard,
	uint32_t serial, struct wl_surface *surface, struct wl_array *keys)
{
	printf("enter\n");
}

/* キーボードフォーカスが離れた時 */

static void _keyboard_leave(void *data, struct wl_keyboard *keyboard,
	uint32_t serial, struct wl_surface *surface)
{
	printf("leave\n");
}

/* キーが押された/離された時 */

static void _keyboard_key(void *data, struct wl_keyboard *keyboard,
	uint32_t serial, uint32_t time, uint32_t key, uint32_t state)
{
	ClientEx *p = (ClientEx *)data;
	xkb_keysym_t sym;

	printf("key | key:%u, %s\n", key,
		(state == WL_KEYBOARD_KEY_STATE_PRESSED)? "press": "release");

	//現在の状態とキーコードから keysym (XKB_KEY_*) を取得

	sym = xkb_state_key_get_one_sym(p->xkb_state, key + 8);

	printf("-- xkb keyshm (0x%X)\n", sym);

	//ESC キーで終了

	if(sym == XKB_KEY_Escape)
		p->b.finish_loop = 1;
}

/* 装飾キーやロックの状態が変化した時 */

static void _keyboard_modifiers(void *data, struct wl_keyboard *keyboard,
	uint32_t serial, uint32_t mods_depressed,
	uint32_t mods_latched, uint32_t mods_locked, uint32_t group)
{
	ClientEx *p = (ClientEx *)data;
	xkb_mod_mask_t mods;

	printf("modifiers | depressed:%u, latched:%u, locaked:%u, group:%u\n",
		mods_depressed, mods_latched, mods_locked, group);

	//状態をセット

	xkb_state_update_mask(p->xkb_state,
		mods_depressed, mods_latched, mods_locked, group, 0, 0);

	//現在の状態のフラグ取得

	mods = xkb_state_serialize_mods(p->xkb_state, XKB_STATE_MODS_EFFECTIVE);

	//状態表示

	printf("-- mods: ");

	if(mods & (1 << xkb_keymap_mod_get_index(p->xkb_keymap, XKB_MOD_NAME_CTRL)))
		printf("Ctrl ");

	if(mods & (1 << xkb_keymap_mod_get_index(p->xkb_keymap, XKB_MOD_NAME_SHIFT)))
		printf("Shift ");

	if(mods & (1 << xkb_keymap_mod_get_index(p->xkb_keymap, XKB_MOD_NAME_ALT)))
		printf("Alt ");

	if(mods & (1 << xkb_keymap_mod_get_index(p->xkb_keymap, XKB_MOD_NAME_LOGO)))
		printf("Logo ");

	if(mods & (1 << xkb_keymap_mod_get_index(p->xkb_keymap, XKB_MOD_NAME_NUM)))
		printf("NumLock ");

	if(mods & (1 << xkb_keymap_mod_get_index(p->xkb_keymap, XKB_MOD_NAME_CAPS)))
		printf("CapsLock ");

	printf("\n");
}

/* キーリピートの情報 */

static void _keyboard_repeat_info(void *data, struct wl_keyboard *keyboard,
	int32_t rate, int32_t delay)
{
	printf("repeat_info | rate:%d, delay:%d\n", rate, delay);
}

static const struct wl_keyboard_listener g_keyboard_listener = {
	_keyboard_keymap, _keyboard_enter, _keyboard_leave, _keyboard_key,
	_keyboard_modifiers, _keyboard_repeat_info
};


//-----------------


/* wl_seat:capabilities */

static void _seat_capabilities(Client *cl,struct wl_seat *seat,uint32_t cap)
{
	ClientEx *p = (ClientEx *)cl;

	if((cap & WL_SEAT_CAPABILITY_KEYBOARD) && !p->keyboard)
	{
		//wl_keyboard 作成

		p->keyboard = wl_seat_get_keyboard(cl->seat);

		wl_keyboard_add_listener(p->keyboard, &g_keyboard_listener, p);
	}
	else if(!(cap & WL_SEAT_CAPABILITY_KEYBOARD) && p->keyboard)
	{
		//wl_keyboard 解放

		_keyboard_release(p);
	}
}


//-----------------


int main(void)
{
	ClientEx *p;
	Window *win;

	p = (ClientEx *)Client_new(sizeof(ClientEx));

	p->b.init_flags = INIT_FLAGS_SEAT;
	p->b.seat_capabilities = _seat_capabilities;

	Client_init(CLIENT(p));

	//

	win = Window_create(CLIENT(p), 256, 256, NULL);

	ImageBuf_fill(win->img, 0xffff0000);
	Window_update(win);

	//

	Client_loop_simple(CLIENT(p));

	//解放

	Window_destroy(win);

	_keyboard_release(p);
	
	Client_destroy(CLIENT(p));

	return 0;
}
