/******************************
 * 出力画面の情報
 ******************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <wayland-client.h>


//------------

typedef struct
{
	struct wl_list link;
	struct wl_output *output;
	uint32_t id,ver;
}output_item;

struct wl_list g_list;

struct wl_display *g_display;
int g_display_callback_cnt = 0;

//------------


//----------------------
// リストデータ
//----------------------


/* 全アイテムを削除 */

static void _list_destroy(void)
{
	output_item *pi,*tmp;

	wl_list_for_each_safe(pi, tmp, &g_list, link)
	{
		if(pi->ver >= WL_OUTPUT_RELEASE_SINCE_VERSION)
			wl_output_release(pi->output);
		else
			wl_output_destroy(pi->output);

		free(pi);
	}
}

/* アイテム追加 */

static void _list_append(uint32_t id,uint32_t ver,struct wl_output *output)
{
	output_item *pi;

	pi = (output_item *)calloc(1, sizeof(output_item));
	if(!pi) return;

	pi->output = output;
	pi->id = id;
	pi->ver = ver;

	wl_list_insert(g_list.prev, &pi->link);
}

/* 指定 ID を削除 */

static void _list_delete(uint32_t id)
{
	output_item *pi;

	wl_list_for_each(pi, &g_list, link)
	{
		if(pi->id == id)
		{
			if(pi->ver >= WL_OUTPUT_RELEASE_SINCE_VERSION)
				wl_output_release(pi->output);
			else
				wl_output_destroy(pi->output);
			
			wl_list_remove(&pi->link);
			free(pi);

			break;
		}
	}
}


//----------------------
// wl_output
//----------------------


static void _output_geometry(void *data, struct wl_output *wl_output,
	int32_t x, int32_t y, int32_t physical_width, int32_t physical_height, int32_t subpixel,
	const char *make, const char *model, int32_t transform)
{
	printf("<geometry>\n"
		"  x:%d, y:%d, physical_width:%d mm, physical_height:%d mm\n"
		"  subpixel:%d, make:'%s', model:'%s', transform:%d\n",
		x, y, physical_width, physical_height,
		subpixel, make, model, transform);
}

static void _output_mode(void *data, struct wl_output *wl_output, uint32_t flags,
	int32_t width, int32_t height, int32_t refresh)
{
	printf("mode | flags:0x%X, width:%d, heigh:%d, refresh:%d mHz\n",
		flags, width, height, refresh);
}

static void _output_done(void *data, struct wl_output *wl_output)
{
	printf("done\n");
}

static void _output_scale(void *data, struct wl_output *wl_output, int32_t factor)
{
	printf("scale | factor:%d\n", factor);
}

static const struct wl_output_listener g_output_listener =
{
	_output_geometry, _output_mode,
	_output_done, _output_scale,
};


//-------------------------
// wl_callback (display)
//-------------------------


static void _display_callback_done(void *data,struct wl_callback *callback,uint32_t time)
{
	wl_callback_destroy(callback);

	g_display_callback_cnt--;

	printf("-- display callback done\n");
}

static const struct wl_callback_listener g_display_callback_listener = {
	_display_callback_done
};


//-------------------------
// sub
//-------------------------


/* イベントの同期コールバック追加 */

static void _add_display_sync(void)
{
	struct wl_callback *callback;

	callback = wl_display_sync(g_display);

	wl_callback_add_listener(callback, &g_display_callback_listener, NULL);

	g_display_callback_cnt++;
}

/* wl_output 追加 */

static void _add_output(struct wl_registry *reg,uint32_t id,uint32_t ver)
{
	struct wl_output *output;

	//output 追加

	output = wl_registry_bind(reg, id, &wl_output_interface, ver);

	_list_append(id, ver, output);

	wl_output_add_listener(output, &g_output_listener, NULL);

	//同期

	_add_display_sync();
}


//---------------------
// wl_registry
//---------------------


static void _registry_global(
	void *data,struct wl_registry *reg,uint32_t id,const char *name,uint32_t ver)
{
	if(strcmp(name, "wl_output") == 0)
		_add_output(reg, id, (ver > 3)? 3: ver);
}

static void _registry_global_remove(void *data,struct wl_registry *registry,uint32_t id)
{
	_list_delete(id);
}

static const struct wl_registry_listener g_reg_listener = {
	_registry_global, _registry_global_remove
};


//----------------


int main(void)
{
	struct wl_registry *reg;

	g_display = wl_display_connect(NULL);
	if(!g_display) return 1;

	//wl_list 初期化

	wl_list_init(&g_list);

	//wl_registry

	reg = wl_display_get_registry(g_display);

	wl_registry_add_listener(reg, &g_reg_listener, NULL);

	_add_display_sync();

	//すべてのイベントが終わるまで待つ

	while(wl_display_dispatch(g_display) != -1
		&& g_display_callback_cnt != 0);

	//

	_list_destroy();

	wl_registry_destroy(reg);

	wl_display_disconnect(g_display);

	return 0;
}
