/******************************
 * xdg-shell 対応クライアント処理
 ******************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <wayland-client.h>

#include "client.h"
#include "client_xdg.h"
#include "imagebuf.h"



//-------------------------
// xdg_surface
//-------------------------


/* 状態変更が確定 */

static void _xdg_surface_configure(
	void *data,struct xdg_surface *surface,uint32_t serial)
{
	WindowXdg *p = (WindowXdg *)data;

	printf("xdg_surface # configure | surface:%p\n", surface);

	xdg_surface_ack_configure(surface, serial);

	//最初の configure イベント時に更新する

	if(p->flag_configure == 0)
	{
		p->flag_configure = 1;

		WindowXdg_update(p);
	}
}

static const struct xdg_surface_listener g_xdg_surface_listener = {
	_xdg_surface_configure
};


//-------------------------
// xdg_toplevel
//-------------------------


/* サイズや状態が変更しようとしている時 */

static void _xdg_toplevel_configure(
	void *data, struct xdg_toplevel *toplevel,
	int32_t width, int32_t height, struct wl_array *states)
{
	uint32_t *ps;

	printf("xdg_toplevel # configure | width:%d, height:%d, states:", width, height);

	wl_array_for_each(ps, states)
	{
		switch(*ps)
		{
			case XDG_TOPLEVEL_STATE_MAXIMIZED:
				printf("MAXIMIZED ");
				break;
			case XDG_TOPLEVEL_STATE_FULLSCREEN:
				printf("FULLSCREEN ");
				break;
			case XDG_TOPLEVEL_STATE_RESIZING:
				printf("RESIZING ");
				break;
			case XDG_TOPLEVEL_STATE_ACTIVATED:
				printf("ACTIVATED ");
				break;
		}
	}

	printf("\n");
}

/* 閉じる */

static void _xdg_toplevel_close(void *data,struct xdg_toplevel *toplevel)
{
	printf("xdg_toplevel # close\n");

	((WindowXdg *)data)->client->b.finish_loop = 1;
}

static const struct xdg_toplevel_listener g_xdg_toplevel_listener = {
	_xdg_toplevel_configure, _xdg_toplevel_close
};


//-------------------------
// xdg_wm_base
//-------------------------


static void _xdg_wm_base_ping(void *data,
	struct xdg_wm_base *xdg_wm_base,uint32_t serial)
{
	//応答する
	xdg_wm_base_pong(xdg_wm_base, serial);
}

const struct xdg_wm_base_listener g_xdg_wm_base_listener = {
	_xdg_wm_base_ping
};


static void _registry_global(void *data,
	struct wl_registry *reg,uint32_t id,const char *name,uint32_t ver)
{
	ClientXdg *p = (ClientXdg *)data;

	if(strcmp(name, "xdg_wm_base") == 0)
	{
		p->xdg.xdg_wm_base = wl_registry_bind(reg, id, &xdg_wm_base_interface, 1);

		xdg_wm_base_add_listener(p->xdg.xdg_wm_base, &g_xdg_wm_base_listener, NULL);
	}

	else if(p->xdg.registry_global_xdg)
		(p->xdg.registry_global_xdg)(data, reg, id, name, ver);
}


//===========================
// ClientXdg
//===========================


/* ClientXdg 作成 */

ClientXdg *ClientXdg_new(int size)
{
	ClientXdg *p;

	if(size < sizeof(ClientXdg))
		size = sizeof(ClientXdg);

	p = (ClientXdg *)Client_new(size);
	if(!p) return NULL;

	p->b.registry_global = _registry_global;

	return p;
}

/* 破棄 */

void ClientXdg_destroy(ClientXdg *p)
{
	if(p)
	{
		if(p->xdg.xdg_wm_base)
			xdg_wm_base_destroy(p->xdg.xdg_wm_base);

		if(p->xdg.cursor_surface)
			wl_surface_destroy(p->xdg.cursor_surface);

		if(p->xdg.cursor_theme)
			wl_cursor_theme_destroy(p->xdg.cursor_theme);
	
		Client_destroy(CLIENT(p));
	}
}

/* 初期化
 *
 * global : 独自に処理するインターフェイスがある場合 (NULL でなし) */

void ClientXdg_init(ClientXdg *p,client_registry_global global)
{
	p->xdg.registry_global_xdg = global;

	Client_init(CLIENT(p));

	//xdg_wm_base がない

	if(!p->xdg.xdg_wm_base)
	{
		printf("[!] not found 'xdg_wm_base'\n");
		ClientXdg_destroy(p);
		exit(1);
	}

	//カーソル

	p->xdg.cursor_theme = wl_cursor_theme_load(NULL, 32, p->b.shm);

	p->xdg.cursor_default = wl_cursor_theme_get_cursor(p->xdg.cursor_theme, "left_ptr");

	if(p->xdg.cursor_default)
	{
		struct wl_cursor_image *img;

		img = p->xdg.cursor_default->images[0];
	
		p->xdg.cursor_surface = wl_compositor_create_surface(p->b.compositor);

		wl_surface_attach(p->xdg.cursor_surface,
			wl_cursor_image_get_buffer(img), 0, 0);

		wl_surface_damage(p->xdg.cursor_surface, 0, 0, img->width, img->height);

		wl_surface_commit(p->xdg.cursor_surface);
	}
}

/* カーソル形状をセット */

void ClientXdg_setCursor(ClientXdg *p,uint32_t serial)
{
	struct wl_cursor_image *img;

	if(p->xdg.cursor_default)
	{
		img = p->xdg.cursor_default->images[0];

		wl_pointer_set_cursor(p->b.pointer, serial, p->xdg.cursor_surface,
			img->hotspot_x, img->hotspot_y);
	}
}


//===========================
// WindowXdg
//===========================


/* トップレベルウィンドウ作成 */

WindowXdg *WindowXdg_createToplevel(ClientXdg *cl,int width,int height)
{
	WindowXdg *p;

	p = (WindowXdg *)calloc(1, sizeof(WindowXdg));
	if(!p) return NULL;

	p->client = cl;

	//wl_surface

	p->surface = wl_compositor_create_surface(cl->b.compositor);

	//xdg_surface

	p->xdg_surface = xdg_wm_base_get_xdg_surface(cl->xdg.xdg_wm_base, p->surface);

	xdg_surface_add_listener(p->xdg_surface, &g_xdg_surface_listener, p);

	//xdg_toplevel

	p->xdg_toplevel = xdg_surface_get_toplevel(p->xdg_surface);
	
	xdg_toplevel_add_listener(p->xdg_toplevel, &g_xdg_toplevel_listener, p);

	//適用

	wl_surface_commit(p->surface);

	//イメージ作成

	p->img = ImageBuf_create(cl->b.shm, width, height);

	return p;
}

/* ウィンドウ破棄 */

void WindowXdg_destroy(WindowXdg *p)
{
	if(p)
	{
		ImageBuf_destroy(p->img);

		xdg_toplevel_destroy(p->xdg_toplevel);

		xdg_surface_destroy(p->xdg_surface);

		wl_surface_destroy(p->surface);
	
		free(p);
	}
}

/* ウィンドウ更新 */

void WindowXdg_update(WindowXdg *p)
{
	wl_surface_attach(p->surface, p->img->buffer, 0, 0);
	wl_surface_damage(p->surface, 0, 0, p->img->width, p->img->height);
	wl_surface_commit(p->surface);
}
