/******************************
 * xdg-shell ポップアップウィンドウ
 *
 * - 左ボタンドラッグで、ウィンドウ位置移動。
 * - 右ボタンクリックで、ポップアップウィンドウ表示。
 *   (クライアントウィンドウ外をクリック、または
 *    ポップアップウィンドウ内を右クリックで閉じる)
 * - 中ボタンで終了。
 ******************************/

#include <stdio.h>
#include <stdlib.h>
#include <linux/input.h>

#include <wayland-client.h>

#include "client.h"
#include "client_xdg.h"
#include "imagebuf.h"


//------------------

//ポップアップウィンドウ

typedef struct
{
	ClientXdg *client;
	struct wl_surface *surface;
	struct xdg_surface *xdg_surface;
	struct xdg_popup *xdg_popup;

	ImageBuf *img;
	WindowXdg *parent;
	int config_w,
		config_h;
	uint8_t flag_configure;
}PopupWindow;


void PopupWindow_destroy(PopupWindow *p);
void PopupWindow_resize(PopupWindow *p,int width,int height);
void PopupWindow_update(PopupWindow *p);

//------------------

//トップレベルウィンドウ
WindowXdg *g_wintop;
//表示中のポップアップウィンドウ
PopupWindow *g_winpop = NULL;
//現在 enter 状態のサーフェス
struct wl_surface *g_on_surface = NULL;
//ポインタ位置
wl_fixed_t g_curx,g_cury;

#define ANCHOR_X  100
#define ANCHOR_Y  100
#define ANCHOR_W  100
#define ANCHOR_H  80

//------------------



//-------------------------
// xdg_surface (popup)
//-------------------------


/* 状態変更が確定 */

static void _xdg_surface_configure(
	void *data,struct xdg_surface *surface,uint32_t serial)
{
	PopupWindow *p = (PopupWindow *)data;

	printf("[popup] xdg_surface # configure | surface:%p\n", surface);

	xdg_surface_ack_configure(surface, serial);

	//ポップアップ時は xdg_popup:configure で渡されたサイズにする

	PopupWindow_resize(p, p->config_w, p->config_h);

	ImageBuf_fill(p->img, 0xff00ff00);

	//最初の configure イベント時に更新する

	if(p->flag_configure == 0)
	{
		p->flag_configure = 1;

		PopupWindow_update(p);
	}
}

static const struct xdg_surface_listener g_xdg_surface_listener = {
	_xdg_surface_configure
};


//-------------------------
// xdg_popup
//-------------------------


/* 位置やサイズ変更時
 *
 * 設定した位置やサイズがサーバーによって調整される場合がある */

static void _popup_configure(void *data, struct xdg_popup *popup,
	int32_t x, int32_t y, int32_t width, int32_t height)
{
	PopupWindow *p = (PopupWindow *)data;

	printf("xdg_popup # configure | popup:%p, (%d,%d)-(%dx%d)\n",
		popup, x, y, width, height);

	p->config_w = width;
	p->config_h = height;
}

/* ポップアップ終了時 */

static void _popup_done(void *data, struct xdg_popup *popup)
{
	PopupWindow *p = (PopupWindow *)data;

	printf("xdg_popup # done | popup:%p\n", popup);

	//

	PopupWindow_destroy(p);

	g_winpop = NULL;
}

static const struct xdg_popup_listener g_popup_listener = {
	_popup_configure, _popup_done
};


//----------------------
// PopupWindow
//----------------------


/* ポップアップ用 xdg_positioner 作成 */

static struct xdg_positioner *_create_popup_positioner(
	PopupWindow *p,int w,int h)
{
	struct xdg_positioner *xdgpos;

	//xdg_positioner 作成

	xdgpos = xdg_wm_base_create_positioner(p->client->xdg.xdg_wm_base);

	//アンカー矩形

	xdg_positioner_set_anchor_rect(xdgpos,
		ANCHOR_X, ANCHOR_Y, ANCHOR_W, ANCHOR_H);

	//ウィンドウサイズ

	xdg_positioner_set_size(xdgpos, w, h);

	//位置をずらす

#if 0
	xdg_positioner_set_offset(xdgpos, 10, 10);
#endif

	//配置

	xdg_positioner_set_anchor(xdgpos, XDG_POSITIONER_ANCHOR_TOP_RIGHT);
	xdg_positioner_set_gravity(xdgpos, XDG_POSITIONER_GRAVITY_BOTTOM_RIGHT);

	//サーバー側で位置やサイズを調整できるようにする

	xdg_positioner_set_constraint_adjustment(xdgpos,
		XDG_POSITIONER_CONSTRAINT_ADJUSTMENT_FLIP_X
		| XDG_POSITIONER_CONSTRAINT_ADJUSTMENT_SLIDE_Y
		| XDG_POSITIONER_CONSTRAINT_ADJUSTMENT_RESIZE_X
		| XDG_POSITIONER_CONSTRAINT_ADJUSTMENT_RESIZE_Y);

	return xdgpos;
}

/* ポップアップウィンドウ作成 */

PopupWindow *PopupWindow_create(ClientXdg *cl,WindowXdg *parent,
	int x,int y,int width,int height,uint32_t serial)
{
	PopupWindow *p;
	struct xdg_positioner *xdgpos;

	p = (PopupWindow *)calloc(1, sizeof(PopupWindow));
	if(!p) return NULL;

	p->client = cl;
	p->parent = parent;

	//wl_surface

	p->surface = wl_compositor_create_surface(cl->b.compositor);

	//xdg_surface

	p->xdg_surface = xdg_wm_base_get_xdg_surface(cl->xdg.xdg_wm_base, p->surface);

	xdg_surface_add_listener(p->xdg_surface, &g_xdg_surface_listener, p);

	//xdg_positioner 作成

	xdgpos = _create_popup_positioner(p, width, height);

	//xdg_popup 作成

	p->xdg_popup = xdg_surface_get_popup(p->xdg_surface,
		parent->xdg_surface, xdgpos);

	xdg_positioner_destroy(xdgpos);

	xdg_popup_add_listener(p->xdg_popup, &g_popup_listener, p);

	//グラブ

	xdg_popup_grab(p->xdg_popup, p->client->b.seat, serial);

	//適用

	wl_surface_commit(p->surface);

	return p;
}

/* ウィンドウ破棄 */

void PopupWindow_destroy(PopupWindow *p)
{
	if(p)
	{
		ImageBuf_destroy(p->img);

		xdg_popup_destroy(p->xdg_popup);

		xdg_surface_destroy(p->xdg_surface);

		wl_surface_destroy(p->surface);
	
		free(p);
	}
}

/* リサイズ */

void PopupWindow_resize(PopupWindow *p,int width,int height)
{
	ImageBuf_destroy(p->img);

	p->img = ImageBuf_create(p->client->b.shm, width, height);
}

/* ウィンドウ更新 */

void PopupWindow_update(PopupWindow *p)
{
	wl_surface_attach(p->surface, p->img->buffer, 0, 0);
	wl_surface_damage(p->surface, 0, 0, p->img->width, p->img->height);
	wl_surface_commit(p->surface);
}


//----------------------
// wl_pointer
//----------------------


/* enter */

static void _pointer_enter(void *data, struct wl_pointer *pointer,
	uint32_t serial, struct wl_surface *surface, wl_fixed_t x, wl_fixed_t y)
{
	printf("wl_pointer # enter | surface:%p\n", surface);

	g_curx = x;
	g_cury = y;
	g_on_surface = surface;

	ClientXdg_setCursor(CLIENT_XDG(data), serial);
}

static void _pointer_leave(void *data, struct wl_pointer *pointer,
	uint32_t serial, struct wl_surface *surface)
{
	printf("wl_pointer # leave | surface:%p\n", surface);

	g_on_surface = NULL;
}

static void _pointer_motion(void *data, struct wl_pointer *pointer,
	uint32_t time, wl_fixed_t x, wl_fixed_t y)
{
	printf("wl_pointer # motion | (%d,%d)\n", x >> 8, y >> 8);

	g_curx = x;
	g_cury = y;
}

/* ボタン */

static void _pointer_button(void *data, struct wl_pointer *pointer,
	uint32_t serial, uint32_t time, uint32_t button, uint32_t state)
{
	ClientXdg *p = CLIENT_XDG(data);

	printf("wl_pointer # button | %s, serial:%u\n",
		(state == WL_POINTER_BUTTON_STATE_PRESSED)? "press":"release",
		serial);

	if(state != WL_POINTER_BUTTON_STATE_PRESSED)
		return;

	//トップレベル上

	if(g_on_surface == g_wintop->surface
		&& !g_winpop)
	{
		switch(button)
		{
			//左ボタンでウィンドウ位置移動
			case BTN_LEFT:
				xdg_toplevel_move(g_wintop->xdg_toplevel, p->b.seat, serial);
				break;
			//右ボタンでポップアップ表示
			case BTN_RIGHT:
				printf("-- create popup | (%d,%d)\n", g_curx >> 8, g_cury >> 8);

				g_winpop = PopupWindow_create(p, g_wintop,
					g_curx >> 8, g_cury >> 8, 200, 300, serial);
				break;
			//中ボタンで終了
			case BTN_MIDDLE:
				p->b.finish_loop = 1;
				break;
		}
	}

	//ポップアップ上

	if(g_winpop && g_on_surface == g_winpop->surface)
	{
		//右ボタンで閉じる
		
		if(button == BTN_RIGHT)
		{
			PopupWindow_destroy(g_winpop);
			g_winpop = NULL;
		}
	}
}

static void _pointer_axis(void *data, struct wl_pointer *pointer,
	uint32_t time, uint32_t axis, wl_fixed_t value)
{

}

static void _pointer_frame(void *data, struct wl_pointer *pointer)
{

}

static void _pointer_axis_source(void *data, struct wl_pointer *pointer, uint32_t axis_source)
{

}

static void _pointer_axis_stop(void *data, struct wl_pointer *pointer, uint32_t time, uint32_t axis)
{

}

static void _pointer_axis_discrete(void *data, struct wl_pointer *pointer, uint32_t axis, int32_t discrete)
{

}

static const struct wl_pointer_listener g_pointer_listener = {
	_pointer_enter, _pointer_leave, _pointer_motion, _pointer_button,
	_pointer_axis, _pointer_frame, _pointer_axis_source,
	_pointer_axis_stop, _pointer_axis_discrete
};


//-----------------


int main(void)
{
	ClientXdg *p;
	WindowXdg *win;

	p = ClientXdg_new(0);

	p->b.init_flags = INIT_FLAGS_SEAT | INIT_FLAGS_POINTER;
	p->b.pointer_listener = &g_pointer_listener;

	ClientXdg_init(p, NULL);

	//ウィンドウ

	g_wintop = win = WindowXdg_createToplevel(p, 256, 256);

	ImageBuf_fill(win->img, 0xffff0000);
	ImageBuf_box(win->img, ANCHOR_X, ANCHOR_Y, ANCHOR_W, ANCHOR_H, 0xff000000);

	//イベントループ

	Client_loop_simple(CLIENT(p));

	//解放

	WindowXdg_destroy(win);

	ClientXdg_destroy(p);

	return 0;
}
